theory ImportPredSet = HOL4ExtraProb+Measure:

types 'a seq = "nat \<Rightarrow> 'a"

consts 
  events:: "('a set set * ('a set \<Rightarrow> real)) \<Rightarrow> 'a set set"
  prob:: "('a set set * ('a set \<Rightarrow> real)) \<Rightarrow> ('a set \<Rightarrow> real)"

translations
  "events" == "Measure.measurable_sets"
  "prob" == "Measure.measure"

constdefs
  prob_space:: "('a set set * ('a set \<Rightarrow> real)) \<Rightarrow> bool"
  "prob_space M \<equiv> (Measure.measure_space M \<and> Measure.measure M {s. True} = 1)"

  bern:: "(bool seq set set) * (bool seq set \<Rightarrow> real)"
  "bern \<equiv> ({S. fst prob.bern (\<lambda>s. s \<in> S)}, (\<lambda>S. snd prob.bern (\<lambda>s. s \<in> S)))"
 (*  SOME bern. prob_space bern \<and> 
  (\<lambda>v. {s. v s} \<in> fst bern) = measure.sigma prob_algebra \<and> (\<forall>s. IN s prob_algebra \<longrightarrow> snd bern {x. s x} = prob_measure s)"
  *)
  P:: "bool seq set \<Rightarrow> real"
  "P \<equiv> prob bern"

  Epsilon:: "bool seq set set" 
  "Epsilon \<equiv> events bern"


lemma sigma_pred: 
  "measure.sigma_algebra a \<Longrightarrow> Sigma_Algebra2.sigma_algebra {M. a (\<lambda>s. s \<in> M)}"
proof -
  assume p: "measure.sigma_algebra a"
  hence cun: "(\<forall>c. (\<exists>f. \<forall>x. c x \<longrightarrow> (\<exists>n::nat. f n = x)) \<and> (\<forall>x. c x \<longrightarrow> a x) \<longrightarrow> a (\<lambda>v. \<exists>p. c p \<and> p v))"
    by (simp add: sigma_algebra_primdef algebra_primdef IN_def COMPL_def UNION_def SUBSET_def 
      GSPEC_DEF_ALT UNIV_def DIFF_def BIGUNION_def EMPTY_def countable_def)
  { fix aa
    assume ai: "\<forall>i::nat. a (\<lambda>s. s \<in> aa i)"
    def c \<equiv> "(\<lambda>p. \<exists>i. (\<lambda>s. s \<in> aa i) = p)"
    { fix p
      assume "c p"
      with c_def have "\<exists>i. (\<lambda>s. s \<in> aa i) = p"
	by simp
    }
    hence "(\<exists>f. \<forall>x. c x \<longrightarrow> (\<exists>n::nat. f n = x))"
      by force   
    also from ai c_def have "(\<forall>x. c x \<longrightarrow> a x)"
      by auto
    ultimately have "a (\<lambda>v. \<exists>p. c p \<and> p v)"
      using cun by simp
    hence "a (\<lambda>v. \<exists>p i. (\<lambda>s. s \<in> aa i) = p \<and> p v)" 
      using c_def by simp
    also have "(\<lambda>v. \<exists>p i. (\<lambda>s. s \<in> aa i) = p \<and> p v)
      = (\<lambda>v. \<exists>p i. p = (\<lambda>s. s \<in> aa i) \<and> p v)"
      by (simp add: eq_sym_conv)
    also 
    have "\<dots> = (\<lambda>v. \<exists>p i. p = (\<lambda>s. s \<in> aa i) \<and> v \<in> aa i)"
    proof (rule ext)
      fix v 
      show "(\<exists>p i. p = (\<lambda>s. s \<in> aa i) \<and> p v) = (\<exists>p i. p = (\<lambda>s. s \<in> aa i) \<and> v \<in> aa i)"
	by force
    qed
    finally have "a (\<lambda>s. \<exists>x. s \<in> aa x)"
      by simp
  }
  hence "\<forall>aa. (\<forall>i::nat. a (\<lambda>s. s \<in> aa i)) \<longrightarrow> a (\<lambda>s. \<exists>x. s \<in> aa x)"
    by simp
  with p show "Sigma_Algebra2.sigma_algebra {M. a (\<lambda>s. s \<in> M)}"
    by (simp add: Sigma_Algebra2.sigma_algebra_def sigma_algebra_primdef algebra_primdef IN_def COMPL_def UNION_def SUBSET_def 
      GSPEC_DEF_ALT UNIV_def DIFF_def BIGUNION_def EMPTY_def countable_def)
qed
  
lemma positive_pred:
  "positive p \<Longrightarrow> Measure.positive ({M. fst p (\<lambda>s. s \<in> M)}, (\<lambda>M. snd p (\<lambda>s. s \<in> M)))"
  by (simp add: POSITIVE_PROB Measure.positive_def measure_primdef measurable_sets_primdef 
    Measure.measurable_sets_def Measure.measure_def EMPTY_def IN_def events_def prob_def split_def)

lemma sum_sumr: "real.sum (0,n) f = sumr 0 n f"
  by (induct n) (simp_all add: sum)

lemma sums_hol:
  "seq.sums f x = f sums x"
  by (auto simp add: sums_def SEQ Series.sums_def LIMSEQ_def sum_sumr real_diff_def) 

lemma countably_additive_pred: assumes ca: "countably_additive p"
  shows "Measure.countably_additive ({M. fst p (\<lambda>s. s \<in> M)}, (\<lambda>M. snd p (\<lambda>s. s \<in> M)))" using ca
proof -
  from ca have step: "\<forall>x. (\<forall>xa. fst p (x xa)) \<and> (\<forall>m n. m \<noteq> n \<longrightarrow> (\<lambda>v. x m v \<and> x n v) = (\<lambda>x. False)) 
    \<and> fst p (\<lambda>v. \<exists>p. (\<exists>xa. p = x xa) \<and> p v) \<longrightarrow> (snd p \<circ> x) sums (snd p (\<lambda>v. \<exists>p. (\<exists>xa. p = x xa) \<and> p v))"
    by (simp add: COUNTABLY_ADDITIVE_PROB IN_def FUNSET_def 
      UNIV_def DISJOINT_def INTER_def IMAGE_def EMPTY_def BIGUNION_def GSPEC_DEF_ALT 
      prob_def events_def 
      measurable_sets_primdef measure_primdef split_def sums_hol)
  { fix f::"nat\<Rightarrow>'a set"
    assume m1: "{y. \<exists>x. y = f x} \<subseteq> {M. fst p (\<lambda>s. s \<in> M)}" 
    and dis: "\<forall>m n. m \<noteq> n \<longrightarrow> f m \<inter> f n = {}" and m2: "fst p (\<lambda>s. \<exists>x. s \<in> f x)"
    def c \<equiv> "(\<lambda>v. (\<lambda>s. s \<in> f v))"
    with m1 have 1: "\<forall>v. fst p (c v)" by blast
    also
    from c_def dis have 2: "(\<forall>m n. m \<noteq> n \<longrightarrow> (\<lambda>v. c m v \<and> c n v) = (\<lambda>x. False))"
    proof -
      { fix m n::nat
	assume ne: "m \<noteq> n"
	{ fix v
	  from dis ne have "(v \<in> f m \<and> v \<in> f n) = False"
	    by blast
	}
	hence "(\<lambda>v. v \<in> f m \<and> v \<in> f n) = (\<lambda>x. False)"
	  by (rule ext)
      }
      thus ?thesis using c_def
	by simp
    qed
    moreover
    have "(\<lambda>s. \<exists>x. s \<in> f x) = (\<lambda>v. \<exists>p. (\<exists>xa. p = c xa) \<and> p v)"
    proof (unfold c_def, rule ext)
      fix s
      show "(\<exists>x. s \<in> f x) = (\<exists>p. (\<exists>xa. p = (\<lambda>s. s \<in> f xa)) \<and> p s)"
	by force
    qed
    ultimately have "(snd p \<circ> c) sums (snd p (\<lambda>s. \<exists>x. s \<in> f x))"
      using step m2 by simp
  }
  thus ?thesis
    by (simp add:  Measure.countably_additive_def Measure.measurable_sets_def Measure.measure_def 
      image_def comp_def) 
qed

lemma bern_bern:
  "prob_space bern \<and> (\<lambda>v. {s. v s} \<in> fst bern) = measure.sigma prob_algebra 
  \<and> (\<forall>s. IN s prob_algebra \<longrightarrow> snd bern  {x. s x} = prob_measure s)"
proof -
  have "probability.prob_space prob.bern"
    by (rule PROB_SPACE_BERN)
  thus ?thesis
   using sigma_pred [of "fst prob.bern"] positive_pred[of "prob.bern"] countably_additive_pred[of "prob.bern"] prob.bern_def
    by (simp add: prob_space_def  Measure.measure_space_def PROB_SPACE Measure.measurable_sets_def 
      measurable_sets_primdef measure_primdef events_def prob_def split_def Measure.measure_def pred_set.UNIV_def bern_def)
qed

  
lemma ps_bern:
  "prob_space bern"
  using bern_bern by simp


lemma UNIV_prob:
  "P {s. True}  = 1"
  using ps_bern P_def
  by (simp add: prob_space_def)

lemma assumes IN: "IN p (probability.events prob.bern)" 
  shows events_pred: "{s. p s} \<in> events bern"
proof -
  from IN have "(measure.sigma prob_algebra) p"
    by (simp add: EVENTS_BERN IN_def)
  also from bern_bern have
    "measure.sigma prob_algebra = (\<lambda>v. {s. v s} \<in> fst bern)"
    by simp
  finally show ?thesis
    by (simp add: Measure.measurable_sets_def)
qed

lemma prob_compl: assumes ps: "prob_space M"
  and ev: "A \<in> events M"
  shows "prob M (-A) = 1-prob M A"
proof -
  have "-A \<union> A = {s. True}"
    by fast
  also
  have disj: "-A \<inter> A = {}"
    by fast
  from ev have "-A \<in> events M" using ps
    by (simp add: prob_space_def Measure.measure_space_def Sigma_Algebra2.sigma_algebra_def)
  with ev disj ps have "prob M (-A \<union> A) = prob M (-A) + prob M A" 
    by (simp add: prob_space_def measure_additive)
  ultimately
  show ?thesis
    using ps 
    by (simp add: prob_space_def) arith
qed



lemma assumes ms: "{s. p s} \<in> events bern"
  shows prob_pred: "prob bern {s. p s} = probability.prob prob.bern p"
  by (simp add: bern_def prob.bern_def prob_def measure_def measure_primdef split_def)
  


end
